import h5py
import numpy as np

from .core import QPImage
from .meta import DATA_KEYS


class IntegrityCheckError(BaseException):
    """Raised when a QPImage data set is incomplete or corrupt"""
    pass


def check(qpi_or_h5file, checks=["attributes", "background"]):
    """Checks various properties of a :class:`qpimage.core.QPImage` instance

    Parameters
    ----------
    qpi_or_h5file: qpimage.core.QPImage or str
        A QPImage object or a path to an hdf5 file
    checks: list of str
        Which checks to perform ("attributes" and/or "background")

    Raises
    ------
    IntegrityCheckError
        if the checks fail
    """
    if isinstance(checks, str):
        checks = [checks]
    for ch in checks:
        if ch not in ["attributes", "background"]:
            raise ValueError("Unknown check: {}".format(check))

    if isinstance(qpi_or_h5file, QPImage):
        qpi = qpi_or_h5file
    else:
        qpi = QPImage(h5file=qpi_or_h5file, h5mode="r")

    # check attributes
    if "attributes" in checks:
        check_attributes(qpi)

    # check background estimation
    if "background" in checks:
        check_background(qpi)


def check_attributes(qpi):
    """Check QPimage attributes

    Parameters
    ----------
    qpi: qpimage.core.QPImage

    Raises
    ------
    IntegrityCheckError
        if the check fails
    """
    missing_attrs = []
    for key in DATA_KEYS:
        if key not in qpi.meta:
            missing_attrs.append(key)
    if missing_attrs:
        msg = "Attributes are missing: {} ".format(missing_attrs) \
            + "in {}!".format(qpi)
        raise IntegrityCheckError(msg)


def check_background(qpi):
    """Check QPimage background data

    Parameters
    ----------
    qpi: qpimage.core.QPImage

    Raises
    ------
    IntegrityCheckError
        if the check fails
    """
    for imdat in [qpi._amp, qpi._pha]:
        try:
            fit, attrs = imdat.get_bg(key="fit", ret_attrs=True)
        except KeyError:
            # No bg correction performed
            pass
        else:
            kwargs = dict(attrs)
            # check if we have a user-defined mask image
            binkey = "estimate_bg_from_mask"
            if binkey in imdat.h5:
                kwargs["from_mask"] = imdat.h5[binkey][:]
            else:
                kwargs["from_mask"] = None
            # compute background correction
            with h5py.File("check.h5",
                           mode="w",
                           driver="core",
                           backing_store=False) as h5:
                # imdat.__class__ is "Amplitude" or "Phase"
                testimdat = imdat.__class__(h5)
                testimdat["raw"] = imdat.raw
                # Set experimental bg data if given
                try:
                    bg = imdat.get_bg("data")
                except KeyError:
                    pass
                else:
                    testimdat.set_bg(bg, key="data")
                # fit bg
                testimdat.estimate_bg(**kwargs)
                # compare
                if not np.allclose(testimdat.get_bg(key="fit"), fit):
                    msg = "Wrong estimated (fitted) background!"
                    raise IntegrityCheckError(msg)



import os
import pathlib
import tempfile

import h5py
import numpy as np
import pytest

import qpimage


def test_series_error_file_is_qpimage():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    qpi1 = qpimage.QPImage(h5file=h5file, h5mode="r")
    tf = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_")
    with qpi1.copy(h5file=tf):
        pass

    try:
        qpimage.QPSeries(qpimage_list=[qpi1], h5file=tf)
    except ValueError:
        pass
    else:
        assert False, "tf is a qpimage file"
    # cleanup
    try:
        os.remove(tf)
    except OSError:
        pass


def test_series_error_key():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    qpi1 = qpimage.QPImage(h5file=h5file, h5mode="r")
    qpi2 = qpi1.copy()

    qps = qpimage.QPSeries(qpimage_list=[qpi1, qpi2])
    try:
        qps.get_qpimage(2)
    except KeyError:
        pass
    else:
        assert False, "get index 2 when length is 2"


def test_series_error_meta():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    qpi1 = qpimage.QPImage(h5file=h5file, h5mode="r")
    qpi2 = qpi1.copy()

    tf = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_")
    with qpimage.QPSeries(qpimage_list=[qpi1, qpi2],
                          h5file=tf,
                          h5mode="a"
                          ):
        pass

    try:
        qpimage.QPSeries(h5file=tf, h5mode="r",
                         meta_data={"wavelength": 550e-9})
    except ValueError:
        pass
    else:
        assert False, "`meta_data` and `h5mode=='r'`"
    # cleanup
    try:
        os.remove(tf)
    except OSError:
        pass


def test_series_error_noqpimage():
    try:
        qpimage.QPSeries(qpimage_list=["hans", 1])
    except ValueError:
        pass
    else:
        assert False, "qpimage list must contain QPImages"


def test_series_from_h5file():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    qpi1 = qpimage.QPImage(h5file=h5file, h5mode="r")
    qpi2 = qpi1.copy()

    tf = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_")
    with qpimage.QPSeries(qpimage_list=[qpi1, qpi2],
                          h5file=tf,
                          h5mode="a"
                          ):
        pass

    qps2 = qpimage.QPSeries(h5file=tf, h5mode="r")
    assert len(qps2) == 2
    assert qps2.get_qpimage(0) == qpi1
    # cleanup
    try:
        os.remove(tf)
    except OSError:
        pass


def test_series_from_list():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    qpi1 = qpimage.QPImage(h5file=h5file, h5mode="r")
    qpi2 = qpi1.copy()

    qps = qpimage.QPSeries(qpimage_list=[qpi1, qpi2])
    assert len(qps) == 2
    assert qps.get_qpimage(0) == qps.get_qpimage(1)


def test_series_h5file():
    tf = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_")
    with h5py.File(tf, mode="a") as fd:
        qps = qpimage.QPSeries(h5file=fd)
        assert len(qps) == 0
    # cleanup
    try:
        os.remove(tf)
    except OSError:
        pass


def test_series_hdf5_hardlink_bg():
    # save compression
    compr = qpimage.image_data.COMPRESSION.copy()
    # disable compression
    qpimage.image_data.COMPRESSION = {}
    # start test
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    bgphase = np.sqrt(np.abs(phase)) + .4

    qpi1 = qpimage.QPImage(data=phase, which_data="phase",
                           bg_data=bgphase, h5dtype="float64")
    qpi2 = qpimage.QPImage(data=phase, which_data="phase", h5dtype="float64")

    tf = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_hardlink_")
    tf = pathlib.Path(tf)
    with qpimage.QPSeries(h5file=tf, h5mode="w") as qps:
        qps.h5.flush()
        s_init = tf.stat().st_size

        qps.add_qpimage(qpi1)
        qps.h5.flush()
        s_bg1 = tf.stat().st_size

        qps.add_qpimage(qpi2, bg_from_idx=0)
        qps.h5.flush()
        s_bg2 = tf.stat().st_size

        qps.add_qpimage(qpi1)
        qps.h5.flush()
        s_bg3 = tf.stat().st_size

        qpih = qps[1].copy()

    assert s_bg2 - s_bg1 < .51 * (s_bg1 - s_init)
    assert s_bg2 - s_bg1 < .51 * (s_bg3 - s_bg2)
    assert np.allclose(qpih.pha, qpi1.pha)
    assert not np.allclose(qpih.pha, qpi2.pha)

    # restore compression
    qpimage.image_data.COMPRESSION = compr


def test_series_meta():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    with pytest.raises((OSError, RuntimeError)):
        qpimage.QPImage(h5file=h5file,
                        meta_data={"wavelength": 333e-9},
                        h5mode="r")

    qpi0 = qpimage.QPImage(h5file=h5file, h5mode="r")

    tf = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_")
    tf = pathlib.Path(tf)
    with qpi0.copy(h5file=tf):
        pass

    qpi1 = qpimage.QPImage(h5file=tf,
                           meta_data={"wavelength": 335e-9})

    assert qpi1.meta["wavelength"] == 335e-9
    qps = qpimage.QPSeries(qpimage_list=[qpi1], meta_data={
                           "wavelength": 400e-9})
    assert qps.get_qpimage(0).meta["wavelength"] == 400e-9


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



import numpy as np
import nrefocus
from scipy.ndimage import gaussian_filter

import pytest

import qpimage


def test_refocus():
    nrkw = {"res": 2,
            "nm": 1,
            "method": "helmholtz",
            "padding": True,
            "d": 5.5}
    meta = {"wavelength": 1e-6,
            "pixel size": 1e-6 / nrkw["res"],
            "medium index": nrkw["nm"]}
    distance = 5.5 * meta["pixel size"]
    size = 40
    x = (np.arange(size) - size / 2).reshape(-1, 1)
    y = (np.arange(size) - size / 2).reshape(1, -1)
    pha = .1
    amp = .5 * (1 + (x**2 + y**2 < size / 3))
    # make smooth to reduce ringing
    amp = gaussian_filter(amp, 1)
    field = amp * np.exp(1j * pha)

    newfield = nrefocus.refocus(field=field, **nrkw)
    qpi0 = qpimage.QPImage(data=newfield,
                           which_data="field",
                           meta_data=meta)

    with pytest.warns(DeprecationWarning, match="kernel"):
        qpi1 = qpi0.refocus(distance=-distance, method=nrkw["method"])

    # sanity
    assert amp.min() < .51
    assert amp.max() > .99
    assert qpi1.amp.min() < .51
    assert qpi1.amp.max() > .99
    assert np.abs(qpi0.pha).max() > 2 * pha

    # refocusing result
    assert not np.allclose(qpi0.amp, amp, rtol=0, atol=8e-4)
    assert not np.allclose(qpi0.pha, pha, rtol=0, atol=3e-4)
    assert np.allclose(qpi1.amp, amp, rtol=0, atol=8e-4)
    assert np.allclose(qpi1.pha, pha, rtol=0, atol=3e-4)


def test_refocus_kernel_different():
    nrkw1 = {"res": 2,
             "nm": 1,
             "padding": True,
             "d": 5.5}
    meta = {"wavelength": 1e-6,
            "pixel size": 1e-6 / nrkw1["res"],
            "medium index": nrkw1["nm"]}
    distance = 5.5 * meta["pixel size"]
    size = 40
    x = (np.arange(size) - size / 2).reshape(-1, 1)
    y = (np.arange(size) - size / 2).reshape(1, -1)
    pha = .1
    amp = .5 * (1 + (x**2 + y**2 < size / 3))
    # make smooth to reduce ringing
    amp = gaussian_filter(amp, 1)
    field = amp * np.exp(1j * pha)

    newfield = nrefocus.refocus(field=field, **nrkw1)
    qpi0 = qpimage.QPImage(data=newfield,
                           which_data="field",
                           meta_data=meta)
    qpi1 = qpi0.refocus(distance=-distance, kernel="helmholtz")
    qpi2 = qpi0.refocus(distance=-distance, kernel="fresnel")
    assert not np.all(qpi1.pha == qpi2.pha)


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



import collections

#: QPImage attributes required for each dataset
DATA_DEF = {
    "medium index": "refractive index of the medium",
    "pixel size": "detector pixel size [m]",
    "time": "acquisition time of the image [s]",
    "wavelength": "imaging wavelength [m]",
    }

#: Optional QPImage attributes
OTHER_DEF = {
    "angle": "tomographic acquisition angle [rad]",
    "date": "acquisition date 'YYYY-MM-DD'",
    "device": "imaging device used",
    "focus": "focus position [m]",
    "identifier": "image identifier",
    "numerical aperture": "numerical aperture",
    "pos x": "x position [m]",
    "pos y": "y position [m]",
    "qpimage version": "qpimage software version used",
    "sim center": "simulation: center of object [px]",
    "sim index": "simulation: refractive index of object",
    "sim model": "simulation: model used",
    "sim radius": "simulation: object radius [m]",
    "software": "imaging software used",
    }

DATA_KEYS = sorted(DATA_DEF.keys())

OTHER_KEYS = sorted(OTHER_DEF.keys())

#: valid :class:`qpimage.core.QPImage` meta data keys
META_KEYS = DATA_KEYS + OTHER_KEYS


class MetaDataMissingError(BaseException):
    """Raised when meta data is missing"""
    pass


class MetaDict(collections.UserDict):
    """Management of meta data variables

    Valid key names are defined in :const:`qpimage.meta.META_KEYS`.
    """
    valid_keys = META_KEYS

    def __init__(self, *args, **kwargs):
        super(MetaDict, self).__init__(*args, **kwargs)
        # check for invalid keys
        for key in self:
            if key not in self.valid_keys:
                raise KeyError("Unknown meta key: '{}'".format(key))

    def __setitem__(self, key, value):
        """Set a meta data key

        The key must be a valid key defined in `self.valid_keys`
        (defaults to :const:`qpimage.meta.META_KEYS`).
        """
        if key not in self.valid_keys:
            raise KeyError("Unknown meta key: '{}'".format(key))
        super(MetaDict, self).__setitem__(key, value)

    def __getitem__(self, *args, **kwargs):
        if args[0] not in self and args[0] in self.valid_keys:
            raise MetaDataMissingError(
                f"No meta data was defined for '{args[0]}'! "
                + "Please make sure you passed the dictionary `meta_data` "
                + "when you loaded your data."
                )
        elif args[0] not in self:
            msg = "Unknown meta key: '{}'!".format(args[0])
            raise KeyError(msg)
        return super(MetaDict, self).__getitem__(*args, **kwargs)



"""Include single scripts with doc string, code, and image

Use case
--------
There is an "examples" directory in the root of a repository,
e.g. 'include_doc_code_img_path = "../examples"' in conf.py
(default). An example is a file ("an_example.py") that consists
of a doc string at the beginning of the file, the example code,
and, optionally, an image file (png, jpg) ("an_example.png").


Configuration
-------------
In conf.py, set the parameter

   fancy_include_path = "../examples"

to wherever the included files reside.


Usage
-----
The directive

   .. fancy_include:: an_example.py

will display the doc string formatted with the first line as a
heading, a code block with line numbers, and the image file.
"""
import io
import os.path as op

from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive
from sphinx.util.nodes import nested_parse_with_titles
from docutils import nodes


class IncludeDirective(Directive):
    required_arguments = 1
    optional_arguments = 0

    def run(self):
        path = self.state.document.settings.env.config.fancy_include_path
        full_path = op.join(path, self.arguments[0])

        with io.open(full_path, "r") as myfile:
            text = myfile.read()

        # add reference
        name = op.basename(full_path)[:-3]
        rst = [".. _example_{}:".format(name),
               "",
               ]

        # add docstring
        source = text.split('"""')
        doc = source[1].split("\n")
        doc.insert(1, "~" * len(doc[0]))  # make title heading

        code = source[2].split("\n")

        for line in doc:
            rst.append(line)

        # image
        for ext in [".png", ".jpg"]:
            image_path = full_path[:-3] + ext
            if op.exists(image_path):
                break
        else:
            image_path = ""
        if image_path:
            rst.append(".. figure:: {}".format(image_path))
            rst.append("")

        # download file
        rst.append(":download:`{}<{}>`".format(
            op.basename(full_path), full_path))

        # code
        rst.append("")
        rst.append(".. code-block:: python")
        rst.append("   :linenos:")
        rst.append("")
        for line in code:
            rst.append("   {}".format(line))
        rst.append("")

        vl = ViewList(rst, "fakefile.rst")
        # Create a node.
        node = nodes.section()
        node.document = self.state.document
        # Parse the rst.
        nested_parse_with_titles(self.state, vl, node)
        return node.children


def setup(app):
    app.add_config_value('fancy_include_path', "../examples", 'html')

    app.add_directive('fancy_include', IncludeDirective)

    return {'version': '0.1'}   # identifies the version of our extension



import warnings

import numpy as np

import qpimage
import qpimage.image_data


def test_bad_keys_error():
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    bgphase = np.sqrt(np.abs(phase))

    qpi = qpimage.QPImage(phase, bg_data=bgphase, which_data="phase")
    clspha = qpi._pha
    try:
        clspha.del_bg("peter")
    except ValueError:
        pass
    else:
        assert False

    try:
        clspha.get_bg("peter")
    except ValueError:
        pass
    else:
        assert False

    try:
        clspha.set_bg(bg=None, key="peter")
    except ValueError:
        pass
    else:
        assert False


def test_del_warning():
    size = 50
    data = np.zeros((size, size), dtype=float)
    qpi = qpimage.QPImage(data=data,
                          which_data="phase",
                          )
    clspha = qpi._pha
    with warnings.catch_warnings(record=True) as rw:
        clspha.del_bg("data")
        assert len(rw) == 1
        msg = str(rw[0].message)
        assert msg.count("data")
        assert msg.lower().count("no bg data")


def test_get_bg():
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    bgphase = np.sqrt(np.abs(phase))

    qpi = qpimage.QPImage(phase, bg_data=bgphase, which_data="phase")
    bgpha = qpi.bg_pha
    clspha = qpi._pha
    assert np.all(bgpha == clspha.bg)
    assert np.all(bgpha == clspha.get_bg(key=None))


def test_get_bg_error():
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    bgphase = np.sqrt(np.abs(phase))

    qpi = qpimage.QPImage(phase, bg_data=bgphase, which_data="phase")
    clspha = qpi._pha
    try:
        clspha.get_bg(key=None, ret_attrs=True)
    except ValueError:
        pass
    else:
        assert False, "attrs of combined bg not supported"


def test_set_bg():
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    bgphase = np.sqrt(np.abs(phase))

    qpi = qpimage.QPImage(phase, bg_data=bgphase, which_data="phase")
    pha = qpi.pha
    clspha = qpi._pha
    assert np.all(pha == clspha.image)
    clspha.set_bg(None, key="data")
    assert not np.all(pha == clspha.image)


def test_set_bg_error():
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    bgphase = np.sqrt(np.abs(phase))

    qpi = qpimage.QPImage(phase, bg_data=bgphase, which_data="phase")
    clspha = qpi._pha
    try:
        clspha.set_bg("bad object", key="data")
    except ValueError:
        pass
    else:
        assert False, "strings cannot be used for bg correction"


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



import numpy as np
from scipy import signal


def find_sideband(ft_data, which=+1, copy=True):
    """Find the side band position of a hologram

    The hologram is Fourier-transformed and the side band
    is determined by finding the maximum amplitude in
    Fourier space.

    Parameters
    ----------
    ft_data: 2d ndarray
        FFt-shifted Fourier transform of the hologram image
    which: +1 or -1
        which sideband to search for:

        - +1: upper half
        - -1: lower half
    copy: bool
        copy `ft_data` before modification

    Returns
    -------
    fsx, fsy : tuple of floats
        coordinates of the side band in Fourier space frequencies
    """
    if copy:
        ft_data = ft_data.copy()

    if which not in [+1, -1]:
        raise ValueError("`which` must be +1 or -1!")

    ox, oy = ft_data.shape
    cx = ox // 2
    cy = oy // 2

    minlo = max(int(np.ceil(ox / 42)), 5)
    if which == +1:
        # remove lower part
        ft_data[cx - minlo:] = 0
    else:
        ft_data[:cx + minlo] = 0

    # remove values around axes
    ft_data[cx - 3:cx + 3, :] = 0
    ft_data[:, cy - 3:cy + 3] = 0

    # find maximum
    am = np.argmax(np.abs(ft_data))
    iy = am % oy
    ix = int((am - iy) / oy)

    fx = np.fft.fftshift(np.fft.fftfreq(ft_data.shape[0]))[ix]
    fy = np.fft.fftshift(np.fft.fftfreq(ft_data.shape[1]))[iy]

    return fx, fy


def fourier2dpad(data, zero_pad=True):
    """Compute the FFT-shifted 2D Fourier transform with zero padding

    Parameters
    ----------
    data: 2d fload ndarray
        real-valued image data
    zero_pad: bool
        perform zero-padding to next order of 2
    """
    if zero_pad:
        # zero padding size is next order of 2
        (N, M) = data.shape
        order = int(max(64., 2**np.ceil(np.log(2 * max(N, M)) / np.log(2))))

        # this is faster than np.pad
        datapad = np.zeros((order, order), dtype=float)
        datapad[:data.shape[0], :data.shape[1]] = data
    else:
        datapad = data

    # Fourier transform
    fft = np.fft.fftshift(np.fft.fft2(datapad))
    return fft


def get_field(hologram, sideband=+1, filter_name="disk", filter_size=1/3,
              filter_size_interpretation="sideband distance",
              subtract_mean=True, zero_pad=True, copy=True, ret_mask=False):
    """Compute the complex field from a hologram using Fourier analysis

    Parameters
    ----------
    hologram: real-valued 2d ndarray
        hologram data (if this is a 3d array, then the first slice
        defined by the first two axes is used)
    sideband: +1, -1, or tuple of (float, float)
        specifies the location of the sideband:

        - +1: sideband in the upper half in Fourier space,
          exact location is found automatically
        - -1: sideband in the lower half in Fourier space,
          exact location is found automatically
        - (float, float): sideband coordinates in
          frequencies in interval [1/"axes size", .5]
    filter_name: str
        specifies the filter to use, one of

        - "disk": binary disk with radius `filter_size`
        - "smooth disk": disk with radius `filter_size` convolved
          with a radial gaussian (`sigma=filter_size/5`)
        - "gauss": radial gaussian (`sigma=0.6*filter_size`)
        - "square": binary square with side length `filter_size`
        - "smooth square": square with side length `filter_size`
          convolved with square gaussian (`sigma=filter_size/5`)
        - "tukey": a square tukey window of width `2*filter_size` and
          `alpha=0.1`
    filter_size: float
        Size of the filter in Fourier space. The interpretation
        of this value depends on `filter_size_interpretation`.
        See `filter_name` for how it is used in filtering.
    filter_size_interpretation: str
        If set to "sideband distance", the filter size is interpreted
        as the relative distance between central band and sideband
        (this is the default). If set to "frequency index", the filter
        size is interpreted as a Fourier frequency index ("pixel size")
        and must be between 0 and `max(hologram.shape)/2`.
    subtract_mean: bool
        If True, remove the mean of the hologram before performing
        the Fourier transform. This setting is recommended as it
        can reduce artifacts from frequencies around the central
        band.
    zero_pad: bool
        Perform zero-padding before applying the FFT. Setting
        `zero_pad` to `False` increases speed but might
        introduce image distortions such as tilts in the phase
        and amplitude data or dark borders in the amplitude data.
    copy: bool
        If set to True, input `data` is not edited.
    ret_mask: bool
        If set to True, return the filter mask used.

    Notes
    -----
    If the input image has three axes, then only the first image
    (defined by the first two axes) is taken (ignore alpha or
    RGB channels).

    The input image is zero-padded as a square image to the next
    order of :math:`2^n`.

    Even though the size of the "gauss" filter approximately matches
    the frequencies of the "disk" filter, it takes into account
    higher frequencies as well and thus suppresses ringing artifacts
    for data that contain jumps in the phase image.
    """
    if len(hologram.shape) == 3:
        # take the first slice (we have alpha or RGB information)
        hologram = hologram[:, :, 0]

    if copy:
        hologram = hologram.astype(dtype=float, copy=True)

    if subtract_mean:
        # remove contributions of the central band
        # (this affects more than one pixel in the FFT
        # because of zero-padding)
        if issubclass(hologram.dtype.type, np.integer):
            hologram = hologram.astype(float)
        hologram -= hologram.mean()

    # Fourier transform
    fft = fourier2dpad(hologram, zero_pad=zero_pad)

    if sideband in [+1, -1]:
        fsx, fsy = find_sideband(fft, which=sideband)
    else:
        fsx, fsy = sideband

    # roll fft such that sideband is located at the image center
    shifted = np.roll(np.roll(fft, -int(fsx * fft.shape[0]), axis=0),
                      -int(fsy * fft.shape[1]), axis=1)

    # coordinates in Fourier space
    assert fft.shape[0] == fft.shape[1]  # square-shaped Fourier domain
    fx = np.fft.fftshift(np.fft.fftfreq(fft.shape[0])).reshape(-1, 1)
    fy = fx.reshape(1, -1)

    if filter_size_interpretation == "sideband distance":
        # filter size based on distance b/w central band and sideband
        if filter_size <= 0 or filter_size >= 1:
            raise ValueError("For sideband distance interpretation, "
                             "`filter_size` must be between 0 and 1; "
                             f"got '{filter_size}'!")
        fsize = np.sqrt(fsx**2 + fsy**2) * filter_size
    elif filter_size_interpretation == "frequency index":
        # filter size given in Fourier index (number of Fourier pixels)
        if filter_size <= 0 or filter_size >= max(hologram.shape)/2:
            raise ValueError("For frequency index interpretation, "
                             "`filter_size` must be between 0 and "
                             f"max(hologram.shape)/2; got '{filter_size}'!")
        # convert to frequencies (compatible with fx and fy)
        fsize = filter_size / fft.shape[0]
    else:
        raise ValueError("Invalid value for `filter_size_interpretation`: "
                         + f"'{filter_size_interpretation}'")

    if filter_name == "disk":
        afilter = (fx**2 + fy**2) <= fsize**2
    elif filter_name == "smooth disk":
        sigma = fsize / 5
        tau = 2 * sigma**2
        radsq = fx**2 + fy**2
        disk = radsq <= fsize**2
        gauss = np.exp(-radsq / tau)
        afilter = signal.convolve(gauss, disk, mode="same")
        afilter /= afilter.max()
    elif filter_name == "gauss":
        sigma = fsize * .6
        tau = 2 * sigma**2
        afilter = np.exp(-(fx**2 + fy**2) / tau)
        afilter /= afilter.max()
    elif filter_name == "square":
        afilter = (np.abs(fx) <= fsize) * (np.abs(fy) <= fsize)
    elif filter_name == "smooth square":
        blur = fsize / 5
        tau = 2 * blur**2
        square = (np.abs(fx) < fsize) * (np.abs(fy) < fsize)
        gauss = np.exp(-(fy**2) / tau) * np.exp(-(fy**2) / tau)
        afilter = signal.convolve(square, gauss, mode="same")
        afilter /= afilter.max()
    elif filter_name == "tukey":
        alpha = 0.1
        rsize = int(min(fx.size, fy.size)*fsize) * 2
        tukey_window_x = signal.tukey(rsize, alpha=alpha).reshape(-1, 1)
        tukey_window_y = signal.tukey(rsize, alpha=alpha).reshape(1, -1)
        tukey = tukey_window_x * tukey_window_y
        afilter = np.zeros(shifted.shape)
        s1 = (np.array(shifted.shape) - rsize)//2
        s2 = (np.array(shifted.shape) + rsize)//2
        afilter[s1[0]:s2[0], s1[1]:s2[1]] = tukey
    else:
        raise ValueError("Unknown filter: {}".format(filter_name))

    # apply filter
    fft_filt = afilter * shifted

    # inverse Fourier transform
    field = np.fft.ifft2(np.fft.ifftshift(fft_filt))
    # crop to original image shape
    cropped = field[:hologram.shape[0], :hologram.shape[1]]

    if ret_mask:
        # shift mask center to sideband location
        shifted_mask = np.roll(
            np.roll(afilter, int(fsx * fft.shape[0]), axis=0),
            int(fsy * fft.shape[1]), axis=1)
        return cropped, shifted_mask
    else:
        return cropped



import os
import tempfile

import h5py
import numpy as np

import qpimage
import qpimage.integrity_check


def test_attributes():
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    bgphase = np.sqrt(np.abs(phase))

    qpi = qpimage.QPImage(phase, bg_data=bgphase, which_data="phase")

    try:
        qpimage.integrity_check.check(qpi, checks="attributes")
    except qpimage.integrity_check.IntegrityCheckError:
        pass
    else:
        assert False, "should raise error, b/c no attributes present"

    qpi2 = qpimage.QPImage(phase, bg_data=bgphase, which_data="phase",
                           meta_data={"medium index": 1.335,
                                      "pixel size": 0.1,
                                      "time": 0.0,
                                      "wavelength": 10e-9,
                                      }
                           )
    qpimage.integrity_check.check(qpi2)


def test_background_mask():
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    bgphase = np.sqrt(np.abs(phase))
    mask = np.zeros_like(bgphase, dtype=bool)
    mask[:10, :] = True
    mask[:, -20] = True
    qpi = qpimage.QPImage(phase, bg_data=bgphase, which_data="phase")
    qpi.compute_bg(which_data="phase",
                   fit_offset="fit",
                   fit_profile="tilt",
                   from_mask=mask)
    qpimage.integrity_check.check(qpi, checks="background")


def test_background_fit():
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    tf = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_")

    with qpimage.QPImage(phase, which_data="phase", h5file=tf) as qpi:
        qpi.compute_bg(which_data="phase",
                       fit_offset="fit",
                       fit_profile="tilt",
                       border_px=5)

    with h5py.File(tf, "a") as h5:
        h5["phase"]["bg_data"]["fit"][:10] = 9

    try:
        qpimage.integrity_check.check(tf, checks="background")
    except qpimage.integrity_check.IntegrityCheckError:
        pass
    else:
        assert False, "wrong bg saved should not work"
    # cleanup
    try:
        os.remove(tf)
    except OSError:
        pass


def test_wrong_check():
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    bgphase = np.sqrt(np.abs(phase))

    qpi = qpimage.QPImage(phase, bg_data=bgphase, which_data="phase")
    try:
        qpimage.integrity_check.check(qpi, checks="peter")
    except ValueError:
        pass
    else:
        assert False, "should raise error, b/c check 'peter' undefined"


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



import numpy as np
import pytest
import qpimage


def hologram(size=64):
    x = np.arange(size).reshape(-1, 1) - size / 2
    y = np.arange(size).reshape(1, -1) - size / 2

    amp = np.linspace(.9, 1.1, size * size).reshape(size, size)
    pha = np.linspace(0, 2, size * size).reshape(size, size)

    rad = x**2 + y**2 > (size / 3)**2
    pha[rad] = 0
    amp[rad] = 1

    # frequencies must match pixel in Fourier space
    kx = 2 * np.pi * -.3
    ky = 2 * np.pi * -.3
    image = (amp**2 + np.sin(kx * x + ky * y + pha) + 1) * 255
    return image


def test_find_sideband():
    size = 40
    ft_data = np.zeros((size, size))
    fx = np.fft.fftshift(np.fft.fftfreq(size))
    ft_data[2, 3] = 1
    ft_data[-3, -2] = 1

    sb1 = qpimage.holo.find_sideband(ft_data=ft_data,
                                     which=+1)
    assert np.allclose(sb1, (fx[2], fx[3]))

    sb2 = qpimage.holo.find_sideband(ft_data=ft_data,
                                     which=-1)
    assert np.allclose(sb2, (fx[-3], fx[-2]))


def test_find_sideband_error():
    size = 40
    ft_data = np.zeros((size, size))
    ft_data[2, 3] = 1
    ft_data[-3, -2] = 1

    try:
        qpimage.holo.find_sideband(ft_data=ft_data,
                                   which=2)
    except ValueError:
        pass
    else:
        assert False, "2 is not a sideband"


def test_fourier2dpad():
    data = np.zeros((100, 120))
    fft1 = qpimage.holo.fourier2dpad(data, zero_pad=True)
    assert fft1.shape == (256, 256)

    fft2 = qpimage.holo.fourier2dpad(data, zero_pad=False)
    assert fft2.shape == data.shape


def test_get_field_error_bad_filter_size():
    holo = hologram()

    with pytest.raises(ValueError, match="must be between 0 and 1"):
        qpimage.holo.get_field(hologram=holo, filter_size=2)


def test_get_field_error_bad_filter_size_interpretation_frequency_index():
    holo = hologram(size=64)

    with pytest.raises(ValueError,
                       match=r"must be between 0 and max\(hologram.shape\)/2"):
        qpimage.holo.get_field(hologram=holo,
                               filter_size_interpretation="frequency index",
                               filter_size=64)


def test_get_field_error_invalid_interpretation():
    holo = hologram()

    with pytest.raises(ValueError,
                       match="Invalid value for `filter_size_interpretation`"):
        qpimage.holo.get_field(hologram=holo,
                               filter_size_interpretation="blequency")


def test_get_field_filter_names():
    holo = hologram()

    kwargs = dict(hologram=holo,
                  sideband=+1,
                  filter_size=1 / 3,
                  subtract_mean=True,
                  zero_pad=True)

    r_disk = qpimage.holo.get_field(filter_name="disk", **kwargs)
    assert np.allclose(
        r_disk[32, 32], 97.307780444912936 - 76.397860381241372j)

    r_smooth_disk = qpimage.holo.get_field(filter_name="smooth disk", **kwargs)
    assert np.allclose(r_smooth_disk[32, 32],
                       108.36665064909741 - 67.176090709644185j)

    r_gauss = qpimage.holo.get_field(filter_name="gauss", **kwargs)
    assert np.allclose(
        r_gauss[32, 32], 108.26984751082375 - 67.116410573093304j)

    r_square = qpimage.holo.get_field(filter_name="square", **kwargs)
    assert np.allclose(
        r_square[32, 32], 102.3285348843612 - 74.139058665601155j)

    r_smsquare = qpimage.holo.get_field(filter_name="smooth square", **kwargs)
    assert np.allclose(
        r_smsquare[32, 32], 105.23157221309754 - 70.593282942004862j)

    r_tukey = qpimage.holo.get_field(filter_name="tukey", **kwargs)
    assert np.allclose(
        r_tukey[32, 32], 113.4826495540899 - 59.546232775481869j)

    try:
        qpimage.holo.get_field(filter_name="unknown", **kwargs)
    except ValueError:
        pass
    else:
        assert False, "unknown filter accepted"


@pytest.mark.parametrize("size", [62, 63, 64])
def test_get_field_interpretation_fourier_index(size):
    """Filter size in Fourier space using Fourier index new in 0.7.0"""
    holo = hologram(size=size)

    ft_data = qpimage.holo.fourier2dpad(data=holo, zero_pad=True)
    fsx, fsy = qpimage.holo.find_sideband(ft_data, which=+1, copy=True)

    kwargs1 = dict(hologram=holo,
                   sideband=+1,
                   filter_name="disk",
                   filter_size=1/3,
                   filter_size_interpretation="sideband distance",
                   subtract_mean=True,
                   zero_pad=True)
    res1 = qpimage.holo.get_field(**kwargs1)

    filter_size_fi = np.sqrt(fsx**2 + fsy**2) / 3 * ft_data.shape[0]
    kwargs2 = dict(hologram=holo,
                   sideband=+1,
                   filter_name="disk",
                   filter_size=filter_size_fi,
                   filter_size_interpretation="frequency index",
                   subtract_mean=True,
                   zero_pad=True)
    res2 = qpimage.holo.get_field(**kwargs2)
    assert np.all(res1 == res2)


@pytest.mark.parametrize("size", [62, 63, 64])
def test_get_field_interpretation_fourier_index_control(size):
    """Filter size in Fourier space using Fourier index new in 0.7.0"""
    holo = hologram(size=size)

    ft_data = qpimage.holo.fourier2dpad(data=holo, zero_pad=True)
    fsx, fsy = qpimage.holo.find_sideband(ft_data, which=+1, copy=True)

    evil_factor = 1.1

    kwargs1 = dict(hologram=holo,
                   sideband=+1,
                   filter_name="disk",
                   filter_size=1/3 * evil_factor,
                   filter_size_interpretation="sideband distance",
                   subtract_mean=True,
                   zero_pad=True)
    res1 = qpimage.holo.get_field(**kwargs1)

    filter_size_fi = np.sqrt(fsx**2 + fsy**2) / 3 * ft_data.shape[0]
    kwargs2 = dict(hologram=holo,
                   sideband=+1,
                   filter_name="disk",
                   filter_size=filter_size_fi,
                   filter_size_interpretation="frequency index",
                   subtract_mean=True,
                   zero_pad=True)
    res2 = qpimage.holo.get_field(**kwargs2)
    assert not np.all(res1 == res2)


@pytest.mark.parametrize("size", [62, 63, 64, 134, 135])
@pytest.mark.parametrize("filter_size", [17, 17.01])
def test_get_field_interpretation_fourier_index_mask_1(size, filter_size):
    """Make sure filter size in Fourier space pixels is correct"""
    holo = hologram(size=size)

    kwargs2 = dict(hologram=holo,
                   sideband=+1,
                   filter_name="disk",
                   filter_size=filter_size,
                   filter_size_interpretation="frequency index",
                   subtract_mean=True,
                   zero_pad=True)
    res, mask = qpimage.holo.get_field(ret_mask=True, **kwargs2)

    # We get 17*2+1, because we measure from the center of Fourier
    # space and a pixel is included if its center is withing the
    # perimeter of the disk.
    assert np.sum(np.sum(mask, axis=0) != 0) == 17*2 + 1


@pytest.mark.parametrize("size", [62, 63, 64, 134, 135])
def test_get_field_interpretation_fourier_index_mask_2(size):
    """Filter size in Fourier space using Fourier index new in 0.7.0"""
    holo = hologram(size=size)

    kwargs2 = dict(hologram=holo,
                   sideband=+1,
                   filter_name="disk",
                   filter_size=16.99,
                   filter_size_interpretation="frequency index",
                   subtract_mean=True,
                   zero_pad=True)
    res, mask = qpimage.holo.get_field(ret_mask=True, **kwargs2)

    # We get two points less than in the previous test, because we
    # loose on each side of the spectrum.
    assert np.sum(np.sum(mask, axis=0) != 0) == 17*2 - 1


def test_get_field_int_copy():
    holo = hologram()
    holo = np.array(holo, dtype=int)

    kwargs = dict(sideband=+1,
                  filter_size=1 / 3,
                  subtract_mean=True,
                  zero_pad=True)

    res1 = qpimage.holo.get_field(hologram=holo,
                                  copy=False,
                                  **kwargs)
    res2 = qpimage.holo.get_field(hologram=holo,
                                  copy=True,
                                  **kwargs)
    res3 = qpimage.holo.get_field(hologram=holo.astype(float),
                                  copy=True,
                                  **kwargs)
    assert np.all(res1 == res2)
    assert np.all(res1 == res3)


def test_get_field_sideband():
    holo = hologram()

    ft_data = qpimage.holo.fourier2dpad(data=holo, zero_pad=True)
    sideband = qpimage.holo.find_sideband(ft_data, which=+1, copy=True)

    kwargs = dict(hologram=holo,
                  filter_name="disk",
                  filter_size=1 / 3,
                  subtract_mean=True,
                  zero_pad=True)
    res1 = qpimage.holo.get_field(sideband=+1, **kwargs)
    res2 = qpimage.holo.get_field(sideband=sideband, **kwargs)
    assert np.all(res1 == res2)


def test_get_field_three_axes():
    holo1 = hologram()
    # create a copy with empty entry in third axis
    holo2 = np.zeros((holo1.shape[0], holo1.shape[1], 2))
    holo2[:, :, 0] = holo1

    kwargs = dict(filter_name="disk",
                  filter_size=1 / 3,
                  subtract_mean=True,
                  zero_pad=True)
    res1 = qpimage.holo.get_field(hologram=holo1, **kwargs)
    res2 = qpimage.holo.get_field(hologram=holo2, **kwargs)
    assert np.all(res1 == res2)


def test_qpimage_holo():
    # create fake hologram
    size = 200
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    kx = -.6
    ky = -.4
    disk_max = 1.5
    # there is a phase disk as data in the hologram
    data = disk_max * ((x - size / 2)**2 + (y - size / 2)**2 < 30**2)
    image = np.sin(kx * x + ky * y + data)
    qpi = qpimage.QPImage(image,
                          which_data="raw-oah",
                          holo_kw={"filter_name": "gauss"})
    qpi.compute_bg(which_data="phase",
                   fit_offset="fit",
                   fit_profile="tilt",
                   border_px=5)
    assert np.allclose(disk_max, qpi.pha.max(), rtol=.01, atol=0)


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



"""Digital hologram of a single cell

This example illustrates how qpimage can be used to analyze
digital holograms. The hologram of a single myeloid leukemia
cell (HL60) shown was recorded using digital holographic microscopy
(DHM). Because the phase-retrieval method used in DHM is based on the
discrete Fourier transform, there always is a residual background
phase tilt which must be removed for further image analysis.
The setup used for recording these data is described in reference
:cite:`Schuermann2015`, which also contains a description of the
hologram-to-phase conversion and phase background correction algorithms
which qpimage is based on.
"""
import matplotlib
import matplotlib.pylab as plt
import numpy as np
import qpimage

# load the experimental data
edata = np.load("./data/hologram_cell.npz")

# create QPImage instance
qpi = qpimage.QPImage(data=edata["data"],
                      bg_data=edata["bg_data"],
                      which_data="raw-oah",
                      # This parameter allows passing arguments to the
                      # hologram-analysis algorithm of qpimage.
                      # (see qpimage.holo.get_field)
                      holo_kw={
                               # For this hologram, the "smooth disk"
                               # filter yields the best trade-off
                               # between interference from the central
                               # band and image resolution.
                               "filter_name": "smooth disk",
                               # Set the filter size to half the distance
                               # between the central band and the sideband.
                               "filter_size": 1/2
                               }
                      )

amp0 = qpi.amp
pha0 = qpi.pha

# background correction
qpi.compute_bg(which_data=["amplitude", "phase"],
               fit_offset="fit",
               fit_profile="tilt",
               border_px=5,
               )

# plot the properties of `qpi`
fig = plt.figure(figsize=(8, 10))

matplotlib.rcParams["image.interpolation"] = "bicubic"
holkw = {"cmap": "gray",
         "vmin": 0,
         "vmax": 200}

ax1 = plt.subplot(321, title="cell hologram")
map1 = ax1.imshow(edata["data"], **holkw)
plt.colorbar(map1, ax=ax1, fraction=.046, pad=0.04)

ax2 = plt.subplot(322, title="bg hologram")
map2 = ax2.imshow(edata["bg_data"], **holkw)
plt.colorbar(map2, ax=ax2, fraction=.046, pad=0.04)

ax3 = plt.subplot(323, title="input phase [rad]")
map3 = ax3.imshow(pha0)
plt.colorbar(map3, ax=ax3, fraction=.046, pad=0.04)

ax4 = plt.subplot(324, title="input amplitude")
map4 = ax4.imshow(amp0, cmap="gray")
plt.colorbar(map4, ax=ax4, fraction=.046, pad=0.04)

ax5 = plt.subplot(325, title="corrected phase [rad]")
map5 = ax5.imshow(qpi.pha)
plt.colorbar(map5, ax=ax5, fraction=.046, pad=0.04)

ax6 = plt.subplot(326, title="corrected amplitude")
map6 = ax6.imshow(qpi.amp, cmap="gray")
plt.colorbar(map6, ax=ax6, fraction=.046, pad=0.04)

# disable axes
[ax.axis("off") for ax in [ax1, ax2, ax3, ax4, ax5, ax6]]

plt.tight_layout()
plt.show()



import os
import pathlib
import tempfile

import h5py
import numpy as np

import qpimage
import qpimage.core


def test_qpimage_copy_to_mem():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    qpi = qpimage.QPImage(h5file=h5file, h5mode="r")
    # create an in-memory copy
    qpi2 = qpi.copy()
    assert np.allclose(qpi.pha, qpi2.pha)
    assert qpi.meta == qpi2.meta


def test_qpimage_copy_to_file():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    qpi = qpimage.QPImage(h5file=h5file, h5mode="r")

    tf = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_")
    with qpi.copy(h5file=tf) as qpi2:
        assert np.allclose(qpi.pha, qpi2.pha)
        assert qpi.meta == qpi2.meta
        assert not np.allclose(qpi2.bg_pha, 0)
        qpi2.clear_bg(which_data="phase", keys="fit")
        assert np.allclose(qpi2.bg_pha, 0)

    with qpimage.QPImage(h5file=tf, h5mode="r") as qpi3:
        assert np.allclose(qpi3.bg_pha, 0)

    # override h5 file
    with h5py.File(tf, mode="a") as h54:
        with qpimage.QPImage(h5file=h54) as qpi4:
            assert np.allclose(qpi4.bg_pha, 0)
        qpi.copy(h5file=h54)
        with qpimage.QPImage(h5file=h54) as qpi5:
            assert not np.allclose(qpi5.bg_pha, 0)

    try:
        os.remove(tf)
    except OSError:
        pass


def test_qpimage_copy_method():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    tf = tempfile.mktemp(suffix=".h5", prefix="qpimage_test_")
    qpimage.core.copyh5(inh5=h5file, outh5=tf)
    qpi1 = qpimage.QPImage(h5file=h5file, h5mode="r")
    qpi2 = qpimage.QPImage(h5file=tf, h5mode="r")
    assert np.allclose(qpi1.pha, qpi2.pha)
    assert qpi1.meta == qpi2.meta

    try:
        os.remove(tf)
    except OSError:
        pass


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



import pathlib

import numpy as np

import qpimage


def test_contains():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    qpi = qpimage.QPImage(h5file=h5file, h5mode="r")
    assert "wavelength" in qpi
    assert "hans-peter" not in qpi


def test_equals():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    qpi = qpimage.QPImage(h5file=h5file, h5mode="r")

    qpi1 = qpi.copy()
    assert qpi1 == qpi

    qpi1["wavelength"] = 123
    assert qpi1 != qpi

    # change phase data
    qpi2 = qpi.copy()
    qpi2.compute_bg(which_data="phase",
                    fit_offset="mean",
                    fit_profile="offset",
                    border_perc=10)

    # change amplitude data
    qpi3 = qpi.copy()
    qpi3.set_bg_data(bg_data=.01 + qpi3.amp, which_data="field")
    assert qpi3 != qpi

    # change identifier (still same data)
    qpi4 = qpi.copy()
    qpi4["identifier"] = "test"
    assert qpi4 == qpi


def test_h5file_confusion():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    try:
        # h5file should be specified with its corresponding parameter
        qpimage.QPImage(h5file, h5mode="r")
    except ValueError:
        pass
    else:
        assert False, "h5file must be given as kwarg!"


def test_info():
    size = 50
    data = np.zeros((size, size), dtype=float)
    mask = np.zeros_like(data, dtype=bool)
    mask[::2, ::2] = True
    qpi = qpimage.QPImage(data=data,
                          which_data="phase",
                          meta_data={"wavelength": 300e-9,
                                     "pixel size": .12e-6,
                                     })
    # mask with border
    qpi.compute_bg(which_data="phase",
                   fit_offset="mean",
                   fit_profile="offset",
                   from_mask=mask,
                   border_px=5)

    info_dict = dict(qpi.info)
    assert 'phase background from mask' in info_dict
    assert 'amplitude background from mask' not in info_dict
    assert info_dict["phase background fit_offset"] == "mean"
    assert info_dict["phase background border_px"] == 5

    qpi.compute_bg(which_data="amplitude",
                   fit_offset="mean",
                   fit_profile="offset",
                   border_px=5)
    info_dict2 = dict(qpi.info)
    assert not info_dict2['amplitude background from mask']


def test_repr():
    # make sure no errors when printing repr
    size = 200
    phase = np.repeat(np.linspace(0, np.pi, size), size)
    phase = phase.reshape(size, size)
    qpi = qpimage.QPImage(phase, which_data="phase",
                          meta_data={"wavelength": 550e-9})
    print(qpi)
    qpi2 = qpimage.QPImage(phase, which_data="phase",
                           meta_data={"wavelength": .1})
    print(qpi2)

    print(qpi._pha)
    print(qpi._amp)


def test_setitem():
    h5file = pathlib.Path(__file__).parent / "data" / "bg_tilt.h5"
    qpi = qpimage.QPImage(h5file=h5file, h5mode="r")

    qpi1 = qpi.copy()
    qpi1["pixel size"] = 0.14e-6

    try:
        qpi1["unknown.data"] = "42"
    except KeyError:
        pass
    else:
        assert False, "Unknown meta name sould raise KeyError."


def test_which_data():
    conv = qpimage.QPImage._conv_which_data
    assert conv("phase") == "phase"
    assert conv("phase,amplitude") == ("phase", "amplitude")
    assert conv("phase, amplitude") == ("phase", "amplitude")
    assert conv("phase,Intensity") == ("phase", "intensity")
    assert conv("Phase") == "phase"
    assert conv("phase,") == "phase"
    assert conv("phase") == "phase"
    assert conv("phase, ") == "phase"
    assert conv(None) is None

    try:
        conv(10)
    except ValueError:
        pass
    else:
        assert False, "which_data can only be list, str, tuple"


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



"""Object-mask background image correction

In some cases, using :ref:`only the border of the phase image
<example_background_poly2o>` for background correction
might not be enough. To increase the area of the background image,
it is possible to mask only the cell area. The :ref:`qpsphere package
<qpsphere:index>` provides the convenience method
:func:`qpsphere.cnvnc.bg_phase_mask_for_qpi` which computes
the background phase mask based on the position and radius of an
automatically detected spherical phase object. The sized of the
mask can be tuned with the `radial_clearance` parameter.

Note that the various methods used in the examples for determining
such a phase mask can be combined. Also note that before
applying the method discussed here, an initial background correction
might be necessary.
"""
import matplotlib.pylab as plt
import numpy as np
# The data are stored in a .jpg file (lossy compression).
# If `PIL` is not found, try installing the `pillow` package.
from PIL import Image
import qpimage
import qpsphere

edata = np.array(Image.open("./data/hologram_cell_curved_bg.jpg"))

# create QPImage instance
qpi = qpimage.QPImage(data=edata,
                      which_data="raw-oah",
                      meta_data={"medium index": 1.335,
                                 "wavelength": 550e-9,
                                 "pixel size": 0.107e-6})
pha0 = qpi.pha

# determine the position of the cell (takes a while)
mask = qpsphere.cnvnc.bg_phase_mask_for_qpi(qpi=qpi,
                                            r0=7e-6,
                                            method="edge",
                                            model="projection",
                                            radial_clearance=1.15)

# background correction using polynomial and mask
qpi.compute_bg(which_data=["phase"],
               fit_offset="fit",
               fit_profile="poly2o",
               from_mask=mask,
               )
pha_corr = qpi.pha

# plot phase data
fig = plt.figure(figsize=(8, 3.3))

phakw = {"cmap": "viridis",
         "interpolation": "bicubic",
         "vmin": pha_corr.min(),
         "vmax": pha_corr.max()}

ax1 = plt.subplot(131, title="input phase")
map1 = ax1.imshow(pha0, **phakw)
plt.colorbar(map1, ax=ax1, fraction=.067, pad=0.04)

ax2 = plt.subplot(132, title="background phase mask")
map2 = ax2.imshow(1.*mask, cmap="gray_r")
plt.colorbar(map2, ax=ax2, fraction=.067, pad=0.04)

ax3 = plt.subplot(133, title="polynomial correction")
map3 = ax3.imshow(pha_corr, **phakw)
plt.colorbar(map3, ax=ax3, fraction=.067, pad=0.04)

# disable axes
[ax.axis("off") for ax in [ax1, ax2, ax3]]

plt.tight_layout(w_pad=0)
plt.show()



from os.path import dirname, realpath, exists
from setuptools import setup, find_packages
import sys


author = u"Paul Müller"
authors = [author]
description = 'library for manipulating quantitative phase images'
name = 'qpimage'
year = "2017"

sys.path.insert(0, realpath(dirname(__file__)) + "/" + name)
from _version import version  # noqa: E402

setup(
    name=name,
    author=author,
    author_email='dev@craban.de',
    url='https://github.com/RI-imaging/qpimage',
    version=version,
    packages=find_packages(),
    package_dir={name: name},
    license="MIT",
    description=description,
    long_description=open('README.rst').read() if exists('README.rst') else '',
    install_requires=["h5py>=2.7.0",
                      "lmfit",
                      "nrefocus[FFTW]>=0.4.3",  # Refocus class
                      "numpy>=1.9.0",
                      "scikit-image>=0.11.0",
                      "scipy>=0.18.0",
                      ],
    python_requires='>=3.6, <4',
    keywords=["digital holographic microscopy",
              "optics",
              "quantitative phase imaging",
              "refractive index",
              "scattering",
              ],
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Science/Research'
                 ],
    platforms=['ALL'],
    )
